#include <stdio.h>
#include <stdlib.h>
#include "chromosome.h"


Chromosome::Chromosome()
{
	alleles = NULL;
}

Chromosome::~Chromosome()
{
	if (alleles != NULL)
		delete[] alleles;
}

void Chromosome::Initialize(int howManyAlleles)
{
	numAlleles = howManyAlleles;
	
	alleles = new allele[numAlleles];

	//alleles are random numbers when we start	
	for (int i = 0; i < numAlleles; i++)
		alleles[i] = random()%256;
}

void Chromosome::PrintAlleleBits(int whichAllele)
{	
	for (int i = (sizeof(allele) * 8) - 1; i >= 0; i--)
	{
		int mask = 1 << i;
		
		if (mask & alleles[whichAllele])
			printf("1");
		else
			printf("0");
	}
	printf(" ");
}

void Chromosome::PrintAllBits()
{
	for (int i = 0; i < numAlleles; i++)
		PrintAlleleBits(i);
}

void Chromosome::PrintAllele(int whichAllele)
{
	printf("%d ", alleles[whichAllele]);
}

void Chromosome::PrintAlleles()
{
	for (int i = 0; i < numAlleles; i++)
	{
		PrintAllele(i);
	}
	printf("\n");
}

unsigned char Chromosome::GetAllele(int whichAllele) 
{ 
	return alleles[whichAllele]; 
};

void Chromosome::SetAllele(int whichAllele, char newAllele) 
{ 
	alleles[whichAllele] = newAllele;
};

void Chromosome::SetAlleles(int howMany, char * newAlleles) 
{ 
	for (int i = 0; i < howMany; i++)
		alleles[i] = newAlleles[i];
};

int Chromosome::GetNumAlleles()
{
	return numAlleles;
}

void Chromosome::Mutate()
{
	//we randomly select an allele and bit to mutute 
	int whichAllele = random()%numAlleles;
	int whichBit = random()%(sizeof(allele) * 8);
	int mask = 1 << whichBit;
	
	alleles[whichAllele] = alleles[whichAllele] ^ mask;
}


/*
this is the most important function!!!
you can stick your own evaluation code in here
return a float between 0.0->1.0
higher is better

this is a pretty naive system: very often you don't know what the best evaluation
would be, so you don't know how to scale things! even if you're doing something
like generating a sinewave and selecting for higher frequencies, how do you know
when you've hit the highest possible frequency? so sometimes you can't just put
out a simple evaluation value. in the frequency example you might return the
actual frequency that has been generated. then you'd have to have a ranking
function that sorts all the evaluations to figure out which is the best. right now
i'm using "closer to 1.0 is better", which is a lot simpler than "higher is better."
the primary difference is that with the former you know when you've hit the global
maximum, whereas with the latter you don't. most "real world" examples are stuck
using the latter, which is what makes figuring out a good evaluation strategy so
important!

i'm using the simple method here just because it's simple and it lets you play
around with a lot of things. 

*/

float Chromosome::Evaluate()
{
//these first two are interesting, as they arrive at the same target (the largest
//possible number) but get there via different routes.


//////////////////////////////////////////////////////////
//this is the classic starter evaluation:
//the more 1's in the chromosome, the better
//////////////////////////////////////////////////////////

	int bitsPerAllele = sizeof(allele) * 8;
	int numBits = numAlleles * bitsPerAllele;
	int numBitsOn = 0;

	for (int i = 0; i < numAlleles; i++)
	{
		for (int j = 0; j < bitsPerAllele; j++)
		{
			int mask = 1 << j;
			if (mask & alleles[i])
				numBitsOn++;
		}
	}		
	
	return (float)numBitsOn/(float)numBits;

//////////////////////////////////////////////////////////


/*
//////////////////////////////////////////////////////////
//this one is really dumb:
//we're just going to add up the alleles and scale to 0.0->1.0
//////////////////////////////////////////////////////////

	float total = 0.0;
	

	for (int i = 0; i < numAlleles; i++)
		total += alleles[i];
	
	total /= 255 * numAlleles;
	
	return total;
	
//////////////////////////////////////////////////////////
*/

/*
//////////////////////////////////////////////////////////
//we can use this one with the html output function
//we'll select for blue (third allele = 255, others = 0)
//
//but imagine that we don't know how color works in html, but 
//we know that if we feed the html parser a set of three two digit
//hex numbers, it'll create a color for us. so we do that, then we
//take a spectrum analyzer and look at the output of the computer
//monitor. we analyze the signal, and the more blue there is in the
//signal, the higher the value we return from this function...
//
//that's more like a real world problem. fortunately, we're not in
//the real world, so we can be lazy!
//////////////////////////////////////////////////////////

	float total = 0.0;
	float temp = 0.0;

	//first red		
	//if it's 0 we'll get 0.0
	//if it's 255 we'll get 1.0
	temp = (float)alleles[0]/255.0;	
	//scale it to 0.0->0.25
	temp *= 0.25;
	//invert
	temp = 0.25 - temp;
	//if red = 0 then we'll be adding .25 to total here
	total += temp;
	
	//again for green
	temp = (float)alleles[1]/255.0;	
	temp *= 0.25;
	temp = 0.25 - temp;
	total += temp;
	
	//now blue
	temp = (float)alleles[2]/255.0;	
	temp *= 0.25;
	//no inversion this time! we want blue!
	total += temp;

	//finally alpha (which we'll ignore later)
	temp = (float)alleles[3]/255.0;	
	temp *= 0.25;
	temp = 0.25 - temp;
	total += temp;
	
	return total;
	
//////////////////////////////////////////////////////////
*/

/*
//////////////////////////////////////////////////////////
//we'll use sound with this one
//each allele generate a frequency component in a signal
//we'll select for octaves
//////////////////////////////////////////////////////////

	float total = 0.0;
	int * freqComponents = new int[numAlleles];

	//we'll scale the components a bit
	//this will give us 0->2550Hz
	for (int i = 0; i < numAlleles; i++)
	{
		freqComponents[i] = alleles[i] * 10;
		if (freqComponents[i] == 0)
			freqComponents[i] = 1;
//		printf("freqComponents[%d]: %d\n",
//			i, freqComponents[i]);
	}

	for (int i = 0; i < numAlleles; i++)
	{
		for (int j = 0; j < numAlleles; j++)
		{
			//we don't want to compare a component to itself!
			if (i != j)
			{
				//see if they divide evenly
				float result = 0.0;
				
				if (freqComponents[i] < freqComponents[j])
					result = (float)freqComponents[j]/(float)freqComponents[i];
				else
					result = (float)freqComponents[i]/(float)freqComponents[j];
				
				if (result == 1.0 || result == 2.0 || result == 4.0)
					total = total + 1;
				
				//if (result == 0.0 || result == 2.0 || result == 3.0 || result == 5.0)
				//	total = total + 1;
				
			}
		}
	}			
	
	total /= (numAlleles * (numAlleles-1));
		
	return total;
	
//////////////////////////////////////////////////////////
*/

}
