#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>

#include "chromosome.h"
#include "evo.h"

#define M_TWO_PI M_PI * 2

Evo::Evo(SettingsStruct * theSettings)
{
	settings = theSettings;	
	chromosomes = NULL;	
}

Evo::~Evo()
{
	Output(1);
	if (chromosomes != NULL)
		delete[] chromosomes;

}

void Evo::Run()
{
	//first we randomize our chromosomes
	InitializeGenePool(settings->numChromosomes, settings->allelesPerChromosome);
	
	//then we just crank through some generations
	for (int i = 0; i < settings->numGenerations; i++)
		DoGeneration();
}
	
void Evo::InitializeGenePool(int numChromosomes, int numAlleles)
{
	chromosomes = new Chromosome[numChromosomes];

	for (int i = 0; i < numChromosomes; i++)
		chromosomes[i].Initialize(numAlleles);
}

void Evo::PrintChromosome(Chromosome * ch)
{
	ch->PrintAlleles();
}

void Evo::DoGeneration()
{
	//the higher the elitism rate, the fewer times we breed
	int howManyToDo = (int)(settings->numChromosomes * (1.0 - settings->elitismRate));
	
	//breed!
	for (int i = 0; i < howManyToDo; i++)
	{
		Chromosome * parentA;
		Chromosome * parentB;
		Chromosome * child = new Chromosome();
		
		child->Initialize(settings->allelesPerChromosome);
		
		//select our two parents
		parentA = SelectParent();	
		parentB = SelectParent();
		
		//cross their genes
		Crossover(parentA, parentB, child);
		
		//possibly mutate the child
		MaybeMutateChromosome(child);	
		
		//we're going to just replace the current worst
		//with the child, which we're hoping will be better
		//you could also replace a random chromosome or ...
		Chromosome * worst = SelectWorst();
		
		for (int i = 0; i < settings->allelesPerChromosome; i++)
			worst->SetAllele(i, child->GetAllele(i));
		
		delete child;
	}
	
	//give us a chance to dump some stats if we want to
	Output(0);
		
}

Chromosome * Evo::SelectParent()
{
	//this is simple tournament selection
	
	//we pick to chromosomes at random
	int a = random()%settings->numChromosomes;
	int b = random()%settings->numChromosomes;
	
	//make sure they're not the same chromosome!
	while (b == a)
		b = random()%settings->numChromosomes;
	
	float evalA = EvaluateChromosome(&chromosomes[a]);
	float evalB = EvaluateChromosome(&chromosomes[b]);
	
	//now select the better of the two as our parent
	if (evalA > evalB)
		return &chromosomes[a];
	else
		return &chromosomes[b];
}


Chromosome * Evo::SelectBest()
{

	float best = 0.0;
	float value;
	int whichIsBest = 0;
	
	Chromosome * currentBest;

	//we simply loop through our chromosomes looking for the best one	
	for (int i = 0; i < settings->numChromosomes; i++)
	{
		value = EvaluateChromosome(&chromosomes[i]);
		if ( value > best)
		{
			currentBest = &chromosomes[i];
			best = value;
			whichIsBest = i;
		}
	}
	return &chromosomes[whichIsBest];
}


Chromosome * Evo::SelectWorst()
{
	float worst = 1.0;
	float value;
	int whichIsWorst = 0;
	
	Chromosome * currentWorst;
	
	//just loop through chromosomes lookingn for the worst one
	for (int i = 0; i < settings->numChromosomes; i++)
	{
		value = EvaluateChromosome(&chromosomes[i]);
		if ( value < worst)
		{
			currentWorst = &chromosomes[i];
			worst = value;
			whichIsWorst = i;
		}
	}
	return &chromosomes[whichIsWorst];
}

/*
we're doing one point crossover here, meaning we just cut
each chromosome at some point and exchange the halves
you can also do multi-point crossover, which allows you to
exchange internal segments
we're cutting only on allele boundaries here.
you can also cut the alleles themselves
*/
void Evo::Crossover(Chromosome * parentA, Chromosome * parentB, Chromosome * child)
{
	int i;
	//select crossover point
	int crossoverAllele = random()%parentA->GetNumAlleles();
	
	//stick new alleles in child
	for (i = 0; i < crossoverAllele; i++)
		child->SetAllele(i, parentA->GetAllele(i));
	for (; i < parentA->GetNumAlleles(); i++)
		child->SetAllele(i, parentB->GetAllele(i));

}

int Evo::MaybeMutateChromosome(Chromosome * ch)
{
	float doit = (float)random()/(float)RAND_MAX;
	
	//higher mutationProbability makes it more likely that we'll mutate
	if (doit <= settings->mutationProbability)
	{
		MutateChromosome(ch);
		return TRUE;
	}
	else
		return FALSE;	
}

void Evo::MutateChromosome(Chromosome * ch)
{
	ch->Mutate();
}

float Evo::EvaluateChromosome(Chromosome * ch)
{
	return ch->Evaluate();
}


/*
you can put some function here that will produce output either to the
terminal or to a file or whatever. in a real implementation you'd want
something more sophisticated than this! 

note that all of these are just outputing rankings -- we're not actually 
doing anything with the expression of the chromosomes (the phenotype). if 
you're expressing the chromosomes as sound (for example) you'd do that in 
the chromosome's evaluation function, not here.

so these functions are mostly just useful for tracking the general fitness 
of a population. this illustrates how generic the simple GA is: you can 
track the fitness of a population without having any knowledge of what the
fitness function is. this is nice because it means you can tweak your 
fitness function like mad without impacting the rest of the algorithm.

if kill == 1 then we will tidy up.
*/

void Evo::Output(int kill)
{

/*
//////////////////////////////////////////////////////////
//print out stats to terminal
//////////////////////////////////////////////////////////

	static int whichGeneration = 0;

	printf("generation %d\n", whichGeneration);
	Chromosome * best = SelectBest();
	Chromosome * worst = SelectWorst();
	printf("best : %f worst: %f\n", EvaluateChromosome(best),
		EvaluateChromosome(worst));
//	printf("best : ");
//	best->PrintAlleles();
	printf("best : ");
	best->PrintAllBits();
	printf("\n");
//	printf("worst: ");
//	worst->PrintAlleles();
	printf("worst: ");
	worst->PrintAllBits();
	printf("\n");
	
	whichGeneration++;
	
//////////////////////////////////////////////////////////
*/

/*
//////////////////////////////////////////////////////////
//print out best chromosome's value to a text file
//////////////////////////////////////////////////////////

	static int firstTime = TRUE;
	static FILE * file;

	if (kill)
	{
		fclose(file);
		return;
	}	

	if (firstTime)
		file = fopen("evo.txt", "w");
	
	firstTime = FALSE;

	fprintf(file, "%f\n", EvaluateChromosome(SelectBest()));
	
//////////////////////////////////////////////////////////
*/

/*
//////////////////////////////////////////////////////////
//use best chromosome value as a phase increment for a sinewave
//write out sine as a raw file for playback by sox via:
// play -c 1 -f s -r 44100 -s w evo.raw
//or convert it to a .wav file with:
// sox -r 44100 -s -w evo.raw evo.wav
//////////////////////////////////////////////////////////
	static int firstTime = TRUE;
	static FILE * file;

	static float phase = 0.0;
	float phaseIncrement = EvaluateChromosome(SelectBest()) * 0.1;

	if (kill)
	{
		fclose(file);
		return;
	}	

	if (firstTime)
		file = fopen("evo.raw", "w");

	firstTime = FALSE;
	
	for (int i = 0; i < 5000; i++)
	{
		float floatOutput = 0.0;

		floatOutput += sin(phase);
		phase += phaseIncrement;
		if (phase > (M_TWO_PI))
			phase -= M_TWO_PI;
		
		int output = (int)(floatOutput * 32767);
	
		fwrite(&output, sizeof(output), 1, file);	
	}

//////////////////////////////////////////////////////////
*/


//////////////////////////////////////////////////////////
//now do all chromosomes at once...
//this one rules!
//////////////////////////////////////////////////////////

	static int firstTime = TRUE;
	static FILE * file;

	static float * phases = new float[settings->numChromosomes];

	if (kill)
	{
		fclose(file);
		delete[] phases;
		return;
	}	

	if (firstTime)
	{
		file = fopen("evo.raw", "w");
		for (int i = 0; i < settings->numChromosomes; i++)
			phases[i] = 0.0;
	}

	firstTime = FALSE;
	
	float * phaseIncrements = new float[settings->numChromosomes];	

	for (int i = 0; i < settings->numChromosomes; i++)
	{
		phaseIncrements[i] = EvaluateChromosome(&chromosomes[i]) * 0.1;
//		printf("phaseIncrements[%d]: %f\n", i, phaseIncrements[i]);
	}
//	printf("\n");
	
	
	for (int i = 0; i < 5000; i++)
	{
		float floatOutput = 0.0;
		for (int j = 0; j < settings->numChromosomes; j++)
		{
			floatOutput += sin(phases[j]);
			phases[j] += phaseIncrements[j];
			if (phases[j] > (M_TWO_PI))
				phases[j] -= M_TWO_PI;
		}
		
		floatOutput /= settings->numChromosomes;
		 
		int output = (int)(floatOutput * 32767);
	
		fwrite(&output, sizeof(output), 1, file);	
	}
	
	delete[] phaseIncrements;
	
//////////////////////////////////////////////////////////


/*
//////////////////////////////////////////////////////////
//this one will produce output in the form of an html table.
//in our evaluation function we're going select for the color
//blue. so here we'll actually express the best chromosome as
//a color.
//
//this is a bit redundant, since the chromosome has already
//been expressed in the evaluation function.
//////////////////////////////////////////////////////////

	static int firstTime = TRUE;
	static FILE * file;
	static int timesThrough = 0;
	
	if (kill)
	{
		fprintf(file, "</table></center></body></html>\n");
		fclose(file);
		return;
	}	

	if (firstTime)
	{
		file = fopen("evo.html", "w");
		fprintf(file, "<html><head><title>evo!!!</title></head>
		<body bgcolor=#000000><center>
		<table><tr>\n");
	}

	if (timesThrough % 10 == 0)
		fprintf(file, "</tr><tr>");
			
	firstTime = FALSE;
	int red = SelectBest()->GetAllele(0);
	int green = SelectBest()->GetAllele(1);
	int blue = SelectBest()->GetAllele(2);
	
	fprintf(file, "<td bgcolor=%.2x%.2x%.2x>%.2x%.2x%.2x</td>\n", red, green, blue,
		red, green, blue);

	
	timesThrough++;

//////////////////////////////////////////////////////////
*/

/*
//////////////////////////////////////////////////////////
//as above, but for all chromosomes
//we'll also mark the best (+) and worst (-)
//////////////////////////////////////////////////////////

	static int firstTime = TRUE;
	static FILE * file;
		
	if (kill)
	{
		fprintf(file, "</table></center></body></html>\n");
		fclose(file);
		return;
	}	

	if (firstTime)
	{
		file = fopen("evo.html", "w");
		fprintf(file, "<html><head><title>evo!!!</title></head>
		<body bgcolor=#000000><center>
		<table><tr>\n");
	}
			
	firstTime = FALSE;
	
	for (int i = 0; i < settings->numChromosomes; i++)
	{
		int red = chromosomes[i].GetAllele(0);
		int green = chromosomes[i].GetAllele(1);
		int blue = chromosomes[i].GetAllele(2);
	
		fprintf(file, "<td align=right bgcolor=%.2x%.2x%.2x>", red, green, blue);
	
		if (&chromosomes[i] == SelectBest())
			fprintf(file, "+");
		else if (&chromosomes[i] == SelectWorst())
			fprintf(file, "-");
	
		fprintf(file, "%.2x%.2x%.2x\n", red, green, blue);
	
		fprintf(file, "</td>");
	
	}

	fprintf(file, "</tr><tr>\n");

//////////////////////////////////////////////////////////
*/

/*
//////////////////////////////////////////////////////////
//this time we'll actually express the genotype as a phenotype
//we'll convert the alleles into frequencies and use them
//to generate sinewaves
//in the evaluation function we're selecting for octaves
// play -c 1 -f s -r 44100 -s w evo.raw
//or convert it to a .wav file with:
// sox -r 44100 -s -w evo.raw evo.wav
//////////////////////////////////////////////////////////
	static int firstTime = TRUE;
	static FILE * file;
	static float * phases = new float[settings->allelesPerChromosome];

	if (kill)
	{
		fclose(file);
		delete[] phases;
		return;
	}	

	if (firstTime)
	{
		file = fopen("evo.raw", "w");
		for (int i = 0; i < settings->allelesPerChromosome; i++)
			phases[i] = 0.0;
	}

	firstTime = FALSE;

	float * phaseIncrements = new float[settings->allelesPerChromosome];
	float frequency;
	
	for (int i = 0; i < settings->allelesPerChromosome; i++)
	{	
		frequency = SelectBest()->GetAllele(i) * 10;
		phaseIncrements[i] = (frequency * M_TWO_PI)/44100.0;
	}

	for (int i = 0; i < 1000; i++)
	{
		float floatOutput = 0.0;
		for (int j = 0; j < settings->allelesPerChromosome; j++)
		{
			floatOutput += sin(phases[j]);
			phases[j] += phaseIncrements[j];
			if (phases[j] > (M_TWO_PI))
				phases[j] -= M_TWO_PI;
		}
		
		floatOutput /= settings->allelesPerChromosome;
		 
		int output = (int)(floatOutput * 32767);
	
		fwrite(&output, sizeof(output), 1, file);	
	}
	
	delete[] phaseIncrements;

//////////////////////////////////////////////////////////
*/

}

int main()
{
	printf("evo starting up...\n");
	printf("reading settings...\n");
	
	srandom(time(NULL));

	SettingsStruct settings;
	ReadSettings(&settings);
	
	Evo theEvo(&settings);
	theEvo.Run();	
	
	printf("evo finished. byebye.\n");

}

void ReadSettings(SettingsStruct * settings)
{

	FILE * file;
	
	file = fopen("settings", "r");
	
	char settingString[16] = "               ";
	int i = 1;
	
	SkipComments(file);
	while(settingString[i-1] != '\n')
	{
		fread(&settingString[i], sizeof(char), 1, file);
		i++;
	}
	settings->allelesPerChromosome = atoi(settingString);
	printf("alleles per chromosome: %d\n", settings->allelesPerChromosome);
	
	SkipComments(file);
	i = 1;
	while(settingString[i-1] != '\n')
	{
		fread(&settingString[i], sizeof(char), 1, file);
		i++;
	}
	settings->numChromosomes = atoi(settingString);
	printf("number of chromosomes: %d\n", settings->numChromosomes);
	
	SkipComments(file);
	i = 1;
	while(settingString[i-1] != '\n')
	{
		fread(&settingString[i], sizeof(char), 1, file);
		i++;
	}
	settings->numGenerations = atoi(settingString);
	printf("number of generations to run: %d\n", settings->numGenerations);

	SkipComments(file);
	i = 1;
	while(settingString[i-1] != '\n')
	{
		fread(&settingString[i], sizeof(char), 1, file);
		i++;
	}
	settings->mutationProbability = (float)atoi(settingString)/100.0;
	printf("mutation probability: %f\n", settings->mutationProbability);

	SkipComments(file);
	i = 1;
	while(settingString[i-1] != '\n')
	{
		fread(&settingString[i], sizeof(char), 1, file);
		i++;
	}
	settings->elitismRate = (float)atoi(settingString) / 100.0;
	printf("elitism rate: %f\n", settings->elitismRate);
	
	fclose(file);

}

void SkipComments(FILE * file)
{
	char ch;
	
	fread(&ch, sizeof(char), 1, file);
	while (ch == '#')
	{
		FindNewLine(file);
		fread(&ch, sizeof(char), 1, file);
	}
	
	fseek(file, -1, SEEK_CUR);
}				

void FindNewLine(FILE * file)
{
	char ch;
	
	while (ch != '\n')
		fread(&ch, sizeof(char), 1, file);
		
}
